/** * 	a.js - a statically compilable javascript library 
 * 
 * copyright Brad Conroy 2013 - unlicensed to the public domain
 * 
 * 
 * Redistribution and use in source form, with or without modification,
 * is permitted provided that the following conditions are met:
 * 
 * * Redistributions of source code must retain the above copyright
 *   and the following disclaimer.
 * 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */


/*	FAQ:
 *	Q. Why do you use bracket notation instead of dot notation?
 *	A. 1.  the closure compiler breaks dot notation, but automatically
 *		changes bracket to dot notation with full optimization
 * 	   2.  it helps to delineate between attributes and methods
 * 
 *	Q. What is with all of the @... comments
 *	A. These are helpers for the closure compiler, for instance the 
 *		@const helper allows removal of if(false) blocks entirely
 * 
 *	Q. What's with all of the const variables.
 *	A. something.toString() can also be represented as:
 *		something['toString'](), so it can also be represented as
 *		/*const/ var toString="toString"; something[toString]()
 * 		this variable can be automatically renamed by the compiler to
 *		var X="toString"; something[X](), which will save 5 bytes for
 *		each occurance, remove the variable if it is not used, or
 *		inline the string if it only occurs once
 * */

/* *
 * TODO:
 * function to combine arrays with separator with pre and post string
 * 	- useful for multiple dataTransfer and query strings "?"+[]&[]&[]""
 *  - or generating tables and lists via html ..."</li><li>"...
 * */

/***********************************************************************
Global Constant Variables
***********************************************************************/
/** F and T are used to minify the many calls for return true/false **/
/** @const */ var F=function(){return!1}
/** @const */ var T=function(){return!0}
/** @const */ var DEBUG=!1
/** @const */ var toLowerCase="toLowerCase"
/** @const */ var toUpperCase="toUpperCase"

if(DEBUG){
/*@const @constructor*/ Object.defineProperty(this, '__stack', {get:function(){var o=Error['prepareStackTrace'],e=new Error,s
    Error['prepareStackTrace']=function(_,s){return s},Error['captureStackTrace'](e, arguments['callee']),s=e['stack']
    Error['prepareStackTrace']=o;return s}})
/*@const @constructor*/ Object.defineProperty(this,'__col__',{get:function(){return __stack[1]['getColumnNumber']()}})
/*@const @constructor*/ Object.defineProperty(this,'__eval_orig__',{get:function(){return __stack[1]['getEvalOrigin']()}})
/*@const @constructor*/ Object.defineProperty(this,'__file__',{get:function(){return __stack[1]['getFileName']()}})
/*@const @constructor*/ Object.defineProperty(this,'__func__',{get:function(){return __stack[1]['getFunctionName']()}})
/*@const @constructor*/ Object.defineProperty(this,'__function__',{get:function(){return __stack[1]['getFunction']()}})
/*@const @constructor*/ Object.defineProperty(this,'__is_constructor__',{get:function(){return __stack[1]['isConstructor']()}})
/*@const @constructor*/ Object.defineProperty(this,'__is_eval__',{get:function(){return __stack[1]['isEval']()}})
/*@const @constructor*/ Object.defineProperty(this,'__is_native__',{get:function(){return __stack[1]['isNative']()}})
/*@const @constructor*/ Object.defineProperty(this,'__is_top_level__',{get:function(){return __stack[1]['isTopLevel']()}})
/*@const @constructor*/ Object.defineProperty(this,'__line__',{get:function(){return __stack[1]['getLineNumber']()}})
/*@const @constructor*/ Object.defineProperty(this,'__method__',{get:function(){return __stack[1]['getMethodName']()}})
/*@const @constructor*/ Object.defineProperty(this,'__this__',{get:function(){return __stack[1]['getThis']()}})
/*@const @constructor*/ Object.defineProperty(this,'__type__',{get:function(){return __stack[1]['getTypeName']()}})
}

/***********************************************************************
Global Functions
* array - taglength,constructor,toString,toLocaleString,join,pop,push,concat,reverse,shift,unshift,slice,splice,sort,filter,forEach,some,every,map,indexOf,lastIndexOf,reduce,reduceRight
* date - constructor,toString,toDateString,toTimeString,toLocaleString,toLocaleDateString,toLocaleTimeString,valueOf,getTime,getFullYear,getUTCFullYear,getMonth,getUTCMonth,getDate,getUTCDate,getDay,getUTCDay,getHours,getUTCHours,getMinutes,getUTCMinutes,getSeconds,getUTCSeconds,getMilliseconds,getUTCMilliseconds,getTimezoneOffset,setTime,setMilliseconds,setUTCMilliseconds,setSeconds,setUTCSeconds,setMinutes,setUTCMinutes,setHours,setUTCHours,setDate,setUTCDate,setMonth,setUTCMonth,setFullYear,setUTCFullYear,toGMTString,toUTCString,getYear,setYear,toISOString,toJSON
* document - createElement,createDocumentFragment,createTextNode,createComment,createCDATASection,createProcessingInstruction,createAttribute,createEntityReference,getElementsByTagName,importNode,createElementNS,createAttributeNS,getElementsByTagNameNS,getElementById,createEvent,createRange,evaluate,execCommand,queryCommandEnabled,queryCommandIndeterm,queryCommandState,queryCommandSupported,queryCommandValue,getElementsByName,elementFromPoint,caretRangeFromPoint,getSelection,getCSSCanvasContext,getElementsByClassName,querySelector,querySelectorAll,webkitGetNamedFlows,adoptNode,createNodeIterator,createTreeWalker,getOverrideStyle,createExpression,createNSResolver,webkitCancelFullScreen,webkitExitFullscreen,webkitExitPointerLock,createTouch,createTouchList,constructor
* event - stopPropagation,preventDefault,initEvent,stopImmediatePropagation,NONE,CAPTURING_PHASE,AT_TARGET,BUBBLING_PHASE,MOUSEDOWN,MOUSEUP,MOUSEOVER,MOUSEOUT,MOUSEMOVE,MOUSEDRAG,CLICK,DBLCLICK,KEYDOWN,KEYUP,KEYPRESS,DRAGDROP,FOCUS,BLUR,SELECT,CHANGE,constructor
* function - length,name,arguments,caller,constructor,bind,toString,call,apply
* math - E,LN10,LN2,LOG2E,LOG10E,PI,SQRT1_2,SQRT2,random,abs,acos,asin,atan,ceil,cos,exp,floor,log,round,sin,sqrt,tan,atan2,pow,max,min
* node - insertBefore,replaceChild,removeChild,appendChild,hasChildNodes,cloneNode,normalize,isSupported,hasAttributes,lookupPrefix,isDefaultNamespace,lookupNamespaceURI,addEventListener,removeEventListener,isSameNode,isEqualNode,compareDocumentPosition,contains,dispatchEvent,ELEMENT_NODE,ATTRIBUTE_NODE,TEXT_NODE,CDATA_SECTION_NODE,ENTITY_REFERENCE_NODE,ENTITY_NODE,PROCESSING_INSTRUCTION_NODE,COMMENT_NODE,DOCUMENT_NODE,DOCUMENT_TYPE_NODE,DOCUMENT_FRAGMENT_NODE,NOTATION_NODE,DOCUMENT_POSITION_DISCONNECTED,DOCUMENT_POSITION_PRECEDING,DOCUMENT_POSITION_FOLLOWING,DOCUMENT_POSITION_CONTAINS,DOCUMENT_POSITION_CONTAINED_BY,DOCUMENT_POSITION_IMPLEMENTATION_SPECIFIC,constructor
* number - constructor,toString,toLocaleString,valueOf,toFixed,toExponential,toPrecision
* object - constructor,toString,toLocaleString,valueOf,hasOwnProperty,isPrototypeOf,propertyIsEnumerable,__defineGetter__,__lookupGetter__,__defineSetter__,__lookupSetter__
* 			prototype,getPrototypeOf,getOwnPropertyDescriptor,keys,defineProperty,defineProperties,create,getOwnPropertyNames,isExtensible,preventExtensions,freeze,isFrozen,seal,isSealed,length,name,arguments,caller
* range - collapse,deleteContents,extractContents,cloneContents,cloneRange,toString,detach,getClientRects,getBoundingClientRect,createContextualFragment,expand,setStart,setEnd,setStartBefore,setStartAfter,setEndBefore,setEndAfter,selectNode,selectNodeContents,compareBoundaryPoints,insertNode,surroundContents,intersectsNode,compareNode,comparePoint,isPointInRange,START_TO_START,START_TO_END,END_TO_END,END_TO_START,NODE_BEFORE,NODE_AFTER,NODE_BEFORE_AND_AFTER,NODE_INSIDE,constructor
* strings - taglength,constructor,valueOf,toString,charAt,charCodeAt,concat,indexOf,lastIndexOf,localeCompare,match,replace,search,slice,split,substring,substr,toLowerCase,toLocaleLowerCase,toUpperCase,toLocaleUpperCase,trim,trimLeft,trimRight,link,anchor,fontcolor,fontsize,big,blink,bold,fixed,italics,small,strike,sub,sup
* window - toString,postMessage,close,blur,focus,ondeviceorientation,ontouchcancel,ontouchend,ontouchmove,ontouchstart,onwebkittransitionend,onwebkitanimationstart,onwebkitanimationiteration,onwebkitanimationend,onsearch,onreset,onwaiting,onvolumechange,onunload,ontimeupdate,onsuspend,onsubmit,onstorage,onstalled,onselect,onseeking,onseeked,onscroll,onresize,onratechange,onprogress,onpopstate,onplaying,onplay,onpause,onpageshow,onpagehide,ononline,onoffline,onmousewheel,onmouseup,onmouseover,onmouseout,onmousemove,onmousedown,onmessage,onloadstart,onloadedmetadata,onloadeddata,onload,onkeyup,onkeypress,onkeydown,oninvalid,oninput,onhashchange,onfocus,onerror,onended,onemptied,ondurationchange,ondrop,ondragstart,ondragover,ondragleave,ondragenter,ondragend,ondrag,ondblclick,oncontextmenu,onclick,onchange,oncanplaythrough,oncanplay,onblur,onbeforeunload,onabort,getSelection,print,stop,open,showModalDialog,alert,confirm,prompt,find,scrollBy,scrollTo,scroll,moveBy,moveTo,resizeBy,resizeTo,matchMedia,setTimeout,clearTimeout,setInterval,clearInterval,requestAnimationFrame,cancelAnimationFrame,webkitRequestAnimationFrame,webkitCancelAnimationFrame,webkitCancelRequestAnimationFrame,atob,btoa,addEventListener,removeEventListener,captureEvents,releaseEvents,getComputedStyle,getMatchedCSSRules,webkitConvertPointFromPageToNode,webkitConvertPointFromNodeToPage,dispatchEvent,webkitRequestFileSystem,webkitResolveLocalFileSystemURL,openDatabase,TEMPORARY,PERSISTENT,constructor
* XMLHttpRequest - open,setRequestHeader,send,abort,getAllResponseHeaders,getResponseHeader,overrideMimeType,addEventListener,removeEventListener,dispatchEvent,UNSENT,OPENED,HEADERS_RECEIVED,LOADING,DONE,constructor
* XMLSerializer - serializeToString,constructor
***********************************************************************/

/** addCommas(string s)
	returns formatted number string with commas
	ex. addCommas("1234567.89"); returns "1,234,567.89" */
function addCommas(s){return s['split'](/(?=(?:\d{3})+(?:\.|$))/g).join( "," );}

/** arrayContains(array a, object o)
	returns 1 if object o is found in array a or 0 if not */
function arrayContains(a,o){var i=a['length']
while(i--){if(a[i]===o)return!0;}
return!1;}

/** array1stMatch(array a, object o)
	returns last instance of object o in array a or -1 if not found */
function array1stMatch(a,o){var l=a['length'],i=0;
while(i<l)if(a[i]===o)return i;
else i++;
return -1;}

/** arrayLastMatch(array a, object o)
	returns last instance of object o in array a or -1 if not found */
function arrayLastMatch(a,o){var i=a['length']
while(i--){if(a[i]===o)break}return i}

function backgroundSetFromSprite(n,w,h,x,y,i){var s=n['style'];
s['background']="url("+i+")";
s['background-position']=x+"px "+y+"px";
s['width']=w;
s['height']=h;
}

/** basename(string s)
	remove ...dir/ and .extension and return resulting string */
function basename(s){s=s.substring(s['lastIndexOf']('/')+1)
var i=s['indexOf']('.')
return s['substring'](0,i!=-1?i:s['length'])}

/** bodyAppend(node n)
	append node n to body */
function bodyAppend(n){document['body']['appendChild'](n)}

/** classAdd(class c,node n)
	adds a class c to node n
	* it will add duplicates, but classReplace/classRemove handle it */
function classAdd(c,n){propertySet(n,"className",n['className']+" "+c)}

/* classAddById(class c,node n)
	adds a class c to node n
	* it will add duplicates, but classReplace/classRemove handle it */
function classAddById(c,id){var n=nodeGetById(id)
propertySet(n,"className",n['className']+" "+c)}

/* add class c to array a of elements */
function classAddAll(c,a){for(i=0;i<a['length'];i++)if(a[i]['className']['indexOf'](c)==-1)classAdd(c,a[i])}

/** classRemove(class c,node n)
	removes a class c from node n
	FIXME pad with " " and replace with " "? */
function classRemove(c,n){classReplace(c," ",n)}

/* remove class c from array a of elements */
function classRemoveAll(c,a){for(i=0;i<a['length'];i++){classRemove(c,a[i]);}}

/** classRemoveById(class c,id id)
	remove class c from 1st node with id
	* see classRemove(class c,id id) */
function classRemoveById(c,id){classRemove(c,nodeGetById(id));}

/** classReplace(class c, class C,node n)
	replace old class c with new class C in node n */
function classReplace(c,C,n){propertySet(n,"className",n['className']['replace'](c,C));}

/* classReplaceById(class c, class C,id id)
	replace old class c with new class C in node with id id */
function classReplaceById(c,C,id){classReplace(c,C,nodeGetById(id));}

/* replace class c with class C in array a of elements */
function classReplaceAll(c,a){for(i=0;i<a['length'];i++){classReplace(c,C,a[i]);}}

/** classSet(class c, node n)
	set class to c for the node n */
function classSet(c,n){n['className']=c;}

/** classSetById(class c,id id)
	set class to c for the id of i */
function classSetById(c,id){classSet(c,nodeGetById(id));}

/** dirname(string s)
	returns string representing the parent directory/ of s */
function dirname(s){return s['substring'](0,s['lastIndexOf']('/')+1);}

/** dragByFunc(node n, function f)
 * 	sets dataTransfer text to value of property p when node n is dragged */
function dragByFunc(n,f){n['draggable']=!0;
n['ondragstart']=function(e){f(n,e);}}

/** dragContents(node n)
	sets dataTransfer text to n.innerHTML */
function dragContents(n){dragPropertyToText(n,'innerHTML');}

/** dragContentsById(id id)
	sets dataTransfer text to n.innerHTML from id id */
function dragContentsById(id){dragContents(nodeGetById(id));}

/** dragPropertyToText(node n, property p)
 * 	sets dataTransfer text to value of property p when node n is dragged */
function dragPropertyToText(n,p){n['draggable']=!0
n['ondragstart']=function(e){e['dataTransfer']['setData']("Text",e['target']?e['target'][p]:e['srcElement'][p])}}

/** dropContentsAdd(node n)
	sets a drop event handler for node n that adds the dataTransfer "Text" to the innerHTML */
function dropContentsAdd(n){n['ondragover']=F;
n['ondrop']=function(a){n['innerHTML']+=a['dataTransfer']['getData']("Text");
return!!0;}}

/** dropContentsAddById(id id)
	sets a drop event handler for node n from id that adds the contents to its innerHTML */
function dropContentsAddById(id){dropContentsAdd(nodeGetById(id));}

/** dropContentsReplace(node n)
	sets a drop event handler for node n that sets its innerHTML to	the dataTransfer "Text" value */
function dropContentsReplace(n){n['ondragover']=F;
n['ondrop']=function(e){n['innerHTML']=e['dataTransfer']['getData']("Text");return!1}}

/** dropContentsReplaceById(id id)
	sets a drop event handler for node from id that replaces the contents of its innerHTML*/
function dropContentsReplaceById(id){dropContentsReplace(nodeGetById(id));}

/** dropTextToFunc(node n, function f)
	sets a drop event handler for node n that uses function f as its handler
	function f will be passed: the target node and the dataTransfer "Text" value
ex. var f=function(t,s){t.style.backgroundColor=s} //sets background the Text value
	* the "Text" is set to the value of property p in the drag event
	see drag(node n, property p) */
function dropTextToFunc(n,f){n['ondragover']=F;
n['ondrop']=function(a){f(n,a['dataTransfer']['getData']("Text"));
return!!0;}}

/** eventAddToId(event ev, id id, function f)
	adds event ev to node with id=id that calls function f 
	see eventAdd(event ev, node n, function f) */
function eventAddToId(ev,id,f){eventAdd(ev,nodeGetById(id),f);}

/** eventsAdd([event,id,function],...)
	adds events ev to nodes with id=id that calls functions f 
	see eventAddToId(event ev, id n, function f) */
function eventsAdd(){var a=arguments,i=a['length'];
while(i--)eventAddToId(a[i][0],a[i][1],a[i][2]);}

/** eventAdd(event ev, node n, function f)
	adds an event ev to node n that calls function f
	does not (currently) return anything
ex.	var f=function(){alert("hello")}
	eventAdd("click",getElementById(id), f) */
function eventAdd(ev,n,f){n["on"+ev]=f;}

/** fontExist(font f)
	returns true if font f is available false otherwise */
function fontExist(f){var s=nodeCreate("span"),ss=s['style'],w;
ss['fontSize']="72px";
ss['visibility']="hidden";
ss['fontFamily']=[(f=="monospace")?'':'monospace','sans-serif','serif'];
s['innerHTML']="MWMWM";
bodyAppend(s);
w=s['offsetWidth'];
ss['fontFamily']=[f,'monospace','sans-serif','serif'];
document['body']['lastChild']=s;
return s['offsetWidth']!=w;}

/** fontList()
	returns an array of supported fonts */
function fontList(){var f="cursive;monospace;serif;sans-serif;fantasy;default;Arial;Arial Black;Arial Narrow;Arial Rounded MT Bold;Bookman Old Style;Bradley Hand ITC;Century;Century Gothic;Comic Sans MS;Courier;Courier New;Georgia;Gentium;Impact;King;Lucida Console;Lalit;Modena;Monotype Corsiva;Papyrus;Tahoma;TeX;Times;Times New Roman;Trebuchet MS;Verdana;Verona"['split'](";"),i=f['length'],a=[];
while(i--)if(fontExist(f[i]))a[a['length']]=f[i];
return a;}

/** langCountry()
	returns string for language and country in the form en-US */
function langCountry(){return(navigator['userLanguage']||navigator['language']);}

/**	lang()
	returns the 2 letter language code
**/
function lang(){return langCountry()['substring'](0,2)}

/**	country()
	returns the country code
**/
function country(){var s=langCountry();
return s['substring'](3,s['length'])[toLowerCase]()}



/** nodesGetByAny(node n, string s)
	finds all matching nodes in parent node 'n'
	that matches any of the selectors in string 's'
	where s is a string of comma separated "tags, #ids, .classes"
	selectors can be combined as "div.hidden, b#title, tag#id.class"
	and returns an array of all elements that match the selectors */
function nodesGetByAny(n,s){return n['querySelectorAll'](s);}

/** nodesGetByTagName(string s)
	returns an array of all nodes of tagname <s> */
function nodesGetByTagName(s){return document['getElementsByTagName'](s);}

/** nodeGetById(id id)
	returns 1st node with id
	* This is only for code size reduction */
function nodeGetById(id){return document['getElementById'](id);}

/** nodeCreate(type t)
	returns a new node of type t
	* This is only for code size reduction */
function nodeCreate(t){return document['createElement'](t);}

/** propertiesSet([id,property,value],...)
	takes a variable number of arrays of type [id,property,value]
	and changes the properties of the matching id to value for each arg
	uses: propertySet(id id, property p, value v) */
function propertiesSet(){var a=arguments,i=a['length'];
while(i--)propertySetById(a[i][0],a[i][1],a[i][2]);}

/** propertySetById(id id, property p, value v)
	changes the property p of the node with id id to value v */
function propertySetById(id,p,v){nodeGetById(id)[p]=v;}

/** propertySet(node n, property p, value v)
	changes the property p of the node n to value v */
function propertySet(n,p,v){n[p]=v;}

/** prependToHead(node n)
	Not adding checks for html without a head tag, seriously? */
function prependToHead(n){var p=nodesGetByTagName('head')[0];
p['insertBefore'](n,p['firstChild']);}

/** prependFromUrl(type t, url u)
	put a tag of type t with source url u right after <head>
	?? application/javascript,css,etc... ?? */
function prependFromUrl(t,u){var n=nodeCreate(t);
n['src']=u;
prependToHead(n);}

/** scriptPrependFromUrl(url u)
	put a script tag with source url right after <head> */
function scriptPrependFromUrl(u){prependFromUrl('script',u);}

/** stylePrependFromUrl(url u)
	put a style tag with source url u right after <head> */
function stylePrependFromUrl(u){prependFromUrl('style',u);}

/** stylePrependFromString(string s)
	put a style tag with coontent of string s right after <head> */
function stylePrependFromString(s){var n=nodeCreate('style');
n['innerHTML']=s;
prependToHead(n);}

function strrchar(s,c){return arrayLastMatch(s,c);}

function strchar(s,c){return array1stMatch(s,c);}

/**	tableFromArray(array a)
 *  returns a table node generated from array a (assumed mxn)
 * 	TODO check me please
 **/
function tableFromArray(a){var w=a.length,h=a[0]['length'],i,j,r,c,t=nodeCreate("table")
for(i=0;i<h;i++){r=nodeCreate("tr")
for(j=0;j<w;j++){c=nodeCreate("td")
c['value']=a[i][j];
r['appendChild'](c);}
t['appendChild'](r);}
return t;}

/**	tooltipSet(node n, string s)
 *	sets the title attribute of node n to string s
 *	most browsers will use this as the tooltip text
 **/
function tooltipSet(n,s){n['title']=s;}

/** urlExists(url u)
	returns true if url "u" exists, otherwise false */
function urlExists(u){var h=XMLHttpRequest();
try{h['open']('HEAD',u,false);
h['send']();
return h['status']!=404;}catch(e){return false;}}

